!+
! Module: emdiag_mod
!
! Authors:  Jean Ballet - SAp  CEA Saclay
!          
! Description: Analyse EPIC-MOS diagnostic images
! 
! Comments: 
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module emdiag_mod

  use emutils_mod
  use edusoft_mod
  use cal
  use dal
  use param  
  use errorHandling

  implicit none 
  
  private 

! Names of the task parameters

    character(len=*), parameter :: inccd      = "diagnosticset",  &
                                   EDUthresh  = "eduthreshold",  &
                                   EMDHlower  = "emdhlowerthreshold",  &
                                   EMDHupper  = "emdhupperthreshold",  &
                                   windowX0   = "windowx0",  &
                                   windowY0   = "windowy0",  &
                                   windowDX   = "windowdx",  &
                                   windowDY   = "windowdy",  &
                                   ksig       = "nsigmas",  &
                                   offset_on  = "offsets",  &
                                   edusoft_on = "findevents",  &
                                   thresh_on  = "cutabovethreshold",  &
                                   subbad     = "maskbadpix",  &
                                   selfov     = "selectfov",  &
                                   inoffvar   = "inoffvarset",  &
                                   writeoff   = "writeoffvarset",  &
                                   outoffvar  = "outoffvarset",  &
                                   writemasked= "writemaskedccdset",  &
                                   maskedfile = "maskedccdset",  &
                                   writeevt   = "writeeventset",  &
                                   outevent   = "eventset",  &
                                   writebri   = "writebrightpixset",  &
                                   brightfile = "brightpixset"

! Name of the extensions

    character(len=*), parameter :: image_ext  = "PRIMARY",  &
                                   event_ext  = "MSIME1",   &
                                   off_ext    = "MSOVE1",  &
                                   offset_ext = "OFFSETS0"

! Keywords to check for compatibility of frame and event files

    character(len=8), dimension(4), parameter :: keycompat = &
         (/"TELESCOP","INSTRUME","CCDID   ","CCDNODE "/)

! Routine name

    character(len=*), parameter :: taskname = "emdiag"

    type(edu_pattern) :: edupat(edu_npat)
    integer(kind=int16), dimension(-1:1,-1:1,0:edu_npat-1) :: patabove

! Global variables
! The window parameters are NOT directly those read as parameters,
! but are those which would have been entered as telecommands
! threshold : EDU threshold
! emdhlow   : EMDH lower threshold
! emdhupp   : EMDH upper threshold
! windX0    : first column in the window (0-609)
! windY0    : first line in the window (0-601)
! windDX    : horizontal width of the window + 2
! windDY    : vertical width of the window + 2
! numsig    : threshold for rejection in OFFSET

    integer(int16) :: threshold
    integer        :: emdhlow, emdhupp, windX0, windY0, windDX, windDY
    real           :: numsig

! Maximum size of the read-out area
! Under/overscans (to be read in CAL)

    integer :: dx1, dx2, dy1, dy2
    integer :: x_size_window, y_size_window

! Pointers to data

    integer(kind=int16), dimension(:,:), pointer :: ccdmap
    integer(kind=int16), dimension(:,:), pointer :: offsetx, offsety

! Function flags
! onoffset    : Should OFFSET  be applied ?
! onedusoft   : Should EDUSOFT be applied ?
! onthresh    : Should THRESH  be applied ?
! onsubbad    : Mask out the CCF bad pixels ?
! onselfov    : Mask out the area outside field of view ?
! writeoffvar : Write the offsets/variance to a file ?
! writemap    : Write the masked CCD map to a file ?
! writeevent  : Write the event list to a file ?
! writebright : Write the map of bright pixels to a file ?

    logical :: onoffset, onedusoft, onthresh, onsubbad, onselfov
    logical :: writeoffvar, writemap, writeevent, writebright

  public emdiag

contains 

  !---------------------------------------------------------------------
  subroutine emdiag
  !---------------------------------------------------------------------

    type(ArrayT)  :: diag_arr           ! DIAGNOSTIC image
    type(TableT)  :: off_tab            ! offset/variance table

    character(len=1024) :: event_out, offvarName, ccd_in, ccd_out, bright_out
    integer :: refX0, refY0
    type(edu_out) :: eduout

! Read parameters

    call getparamvalues(ccd_in, ccd_out, offvarName, event_out, bright_out)

    call readmap(ccd_in, diag_arr, refX0, refY0)

    if (onoffset) then
       call getOffsets(refX0, refY0, offvarName, diag_arr, off_tab)
    else
       call getoffvar(offvarName, diag_arr, off_tab)
    endif

    if (onedusoft)  &
       call educall(refX0, refY0, event_out, diag_arr, off_tab, eduout)

    if (onthresh)   &
       call thresh(refX0, refY0, bright_out, diag_arr, off_tab, eduout)

    if (writemap)   &
       call putmap(refX0, refY0, ccd_out, diag_arr, off_tab)

    call release(parent(off_tab))
    call release(parent(diag_arr))

  end subroutine emdiag

! Compute the offsets

  subroutine getOffsets(refX0, refY0, offvar_out, diag_arr, off_tab)

! ref..      : Window parameters of the input map
! offvar_out : Name of the output offset/variance file
! diag_arr   : Handle to the CCD map array
! off_tab    : Handle to the offset/variance table

    integer,          intent(in)  :: refX0, refY0
    character(len=*), intent(in)  :: offvar_out
    type(ArrayT),     intent(in)  :: diag_arr
    type(TableT),     intent(out) :: off_tab

    type(DataSetT) :: off_set
    type(ColumnT)  :: offx_col, offy_col
    integer(kind=int16), dimension(:,:), pointer :: fovMap
    integer(kind=int16), dimension(:,:), allocatable :: extimg
    integer, dimension(16) :: border
    integer, dimension(:), allocatable :: numpixrow, totrow
    integer, dimension(:), allocatable :: numpixcol, totcol
    integer :: ngoodpix, nbadpix, ndof, noffsets
    integer :: windX1, windY1, xref0, yref0, xref1, yref1, ir , jr
    integer(kind=int16) :: xwin0, xwin1, ywin0, ywin1
    integer :: xrel0, xrel1, yrel0, yrel1
    real(double) :: averim, variance

    if (writeoffvar) then
       call message(AppMsg,Verbose,"Opening the offset/variance file ...")
       off_set = dataSet(offvar_out, Create)
    else
       off_set = dataSet('emdiag.temp', Temp)
    endif

    off_tab = addTable(off_set, off_ext, 1, "Offset/variance data")
    call setAttributes(attributable(off_tab), attributable(diag_arr))
    call setAttribute(off_tab, "DATATYPE", "OFFVAR.EL")
    call setAttribute(off_tab, "CCDMAP", trim(name(parent(diag_arr))),  &
                      comment="Input diagnostic file")
    call setAttribute(off_tab, ksig, numsig,  &
                      comment="Threshold for pixel rejection")
    call setAttribute(off_tab, "WINDOWX0", windX0,  &
                      comment="X-coordinate of bottom left corner of window")
    call setAttribute(off_tab, "WINDOWY0", windY0,  &
                      comment="Y-coordinate of bottom left corner of window")
    call setAttribute(off_tab, "WINDOWDX", windDX,  &
                      comment="Size, along X-axis, of window")
    call setAttribute(off_tab, "WINDOWDY", windDY,  &
                      comment="Size, along Y-axis, of window")
    call setAttribute(off_tab, "EDUTHR", 0)  ! Expected in ODF ICD
    if (onsubbad)  call addComment(off_tab,  &
                                   "Bad pixels from the CCF subtracted")
    offx_col = addColumn(off_tab, "COLOFFST", Integer16,  &
               comment="Row offset value", dimensions=(/x_size_window/))
    offy_col = addColumn(off_tab, "ROWOFFST", Integer16,  &
               comment="Column offset value", dimensions=(/y_size_window/))
    offsetx => int16VectorData(offx_col)
    offsety => int16VectorData(offy_col)
    offsetx = 0
    offsety = 0

    call message(AppMsg,Verbose,"Computing the offsets ...")

! Allocate arrays

    allocate(extimg(-2:x_size_window+1,-2:y_size_window+1))
    allocate(numpixrow(0:x_size_window-1),totrow(0:x_size_window-1))
    allocate(numpixcol(0:y_size_window-1),totcol(0:y_size_window-1))

    numpixrow = 0
    totrow = 0
    numpixcol = 0
    totcol = 0
    ngoodpix = 0
    nbadpix = 0

    windX1 = windX0 + windDX - 1
    windY1 = windY0 + windDY - 1
    xref0 = windX0 - refX0
    xref1 = windX1 - refX0
    yref0 = windY0 - refY0
    yref1 = windY1 - refY0

! extimg has two border lines/columns on each side

    extimg(windX0:windX1,windY0:windY1) = ccdmap(xref0:xref1,yref0:yref1)
    extimg(windX0-1,windY0:windY1) = extimg(windX0,windY0:windY1)
    extimg(windX0-2,windY0:windY1) = extimg(windX0,windY0:windY1)
    extimg(windX1+1,windY0:windY1) = extimg(windX1,windY0:windY1)
    extimg(windX1+2,windY0:windY1) = extimg(windX1,windY0:windY1)
    extimg(windX0-2:windX1+2,windY0-1) = extimg(windX0-2:windX1+2,windY0)
    extimg(windX0-2:windX1+2,windY0-2) = extimg(windX0-2:windX1+2,windY0)
    extimg(windX0-2:windX1+2,windY1+1) = extimg(windX0-2:windX1+2,windY1)
    extimg(windX0-2:windX1+2,windY1+2) = extimg(windX0-2:windX1+2,windY1)

! Remove abnormal pixels from computation by test on difference with median
! border contains all pixels in outer corona around current pixel

    do jr = windY0, windY1
       do ir = windX0, windX1
          border(1:4)   = extimg(ir-2:ir+1,jr-2)
          border(5:8)   = extimg(ir+2,jr-2:jr+1)
          border(9:12)  = extimg(ir-1:ir+2,jr+2)
          border(13:16) = extimg(ir-2,jr-1:jr+2)
          if ( test_valid(extimg(ir,jr),border,numsig) ) then
             numpixrow(ir) = numpixrow(ir) + 1
             numpixcol(jr) = numpixcol(jr) + 1
             totrow(ir) = totrow(ir) + extimg(ir,jr)
             totcol(jr) = totcol(jr) + extimg(ir,jr)
             ngoodpix = ngoodpix + 1
          else
             ccdmap(ir-refX0,jr-refY0) = 0   ! used in variance computation
             nbadpix = nbadpix + 1
          endif
       enddo
    enddo

    if (ngoodpix < 2) then
       write(errstr,*) "the overall number of good pixels", ngoodpix,  &
                       " is less than 2."
       call error("getOffsets01")
    endif
    if (minval(numpixrow(windX0:windX1)) == 0) then
       write(errstr,*) "no good pixel has been found in some columns. ",  &
                       "The corresponding offsets are set to 0."
       call warning("getOffsets10")
    endif
    if (minval(numpixcol(windY0:windY1)) == 0) then
       write(errstr,*) "no good pixel has been found in some rows. ",  &
                       "The corresponding offsets are set to 0."
       call warning("getOffsets10")
    endif
    ndof = ngoodpix - windDX - windDY + 1
    if (ndof < 1) then
       write(errstr,*) "The number of good pixels", ngoodpix,  &
          " is less than the number of offsets", windDX+windDY-1,  &
          ". The result is not reliable."
       call warning("getOffsets12")
       ndof = 1 
    endif
    averim = sum(real(totrow,kind=double)) / ngoodpix
    write(errstr,*) "Average value over the window =", averim
    call message(AppMsg,Verbose)
    where (numpixrow > 0)
       offsetx(0:x_size_window-1,0) = int(real(totrow)/numpixrow  &
                                          -averim/2+0.5,kind=int16)
    endwhere
    where (numpixcol > 0)
       offsety(0:y_size_window-1,0) = int(real(totcol)/numpixcol  &
                                          -averim/2+0.5,kind=int16)
    endwhere

    write(errstr,*) ngoodpix, " pixels used for computing the offsets"
    call message(AppMsg,Verbose)
    write(errstr,*) nbadpix, " pixels rejected for computing the offsets"
    call message(AppMsg,Verbose)

! Second run for the area within the field of view

    if (onselfov) then
       call message(AppMsg,Sparse,"Rejecting pixels out of the field of view")
       xwin0 = max(windX0,dx1) - dx1 + 1
       xwin1 = min(windX0+windDX,x_size_window-dx2) - dx1 + 1
       ywin0 = max(windY0,dy1) - dy1 + 1
       ywin1 = min(windY0+windDY,y_size_window-dy2) - dy1 + 1
       xrel0 = xwin0 - 1 + dx1 - refX0
       xrel1 = xwin1 - 2 + dx1 - refX0
       yrel0 = ywin0 - 1 + dy1 - refY0
       yrel1 = ywin1 - 2 + dy1 - refY0
       call CAL_getFOVmap(fovMap,xwin0,ywin0,xwin1-xwin0,ywin1-ywin0)
       where (fovMap == 0)
          ccdmap(xrel0:xrel1,yrel0:yrel1) = 0
       endwhere
       call CAL_releaseMemory(fovMap)

       numpixrow = 0
       totrow = 0
       numpixcol = 0
       totcol = 0
       ngoodpix = 0
       nbadpix = 0
       do jr = yrel0, yrel1
          do ir = xrel0, xrel1
             if (ccdmap(ir,jr) > 0) then
                numpixrow(ir+refX0) = numpixrow(ir+refX0) + 1
                numpixcol(jr+refY0) = numpixcol(jr+refY0) + 1
                totrow(ir+refX0) = totrow(ir+refX0) + ccdmap(ir,jr)
                totcol(jr+refY0) = totcol(jr+refY0) + ccdmap(ir,jr)
                ngoodpix = ngoodpix + 1
             else
                nbadpix = nbadpix + 1
             endif
          enddo
       enddo

       if (ngoodpix < 2) then
          write(errstr,*) "the overall number of good pixels in the FOV",  &
                          ngoodpix, " is less than 2."
          call error("getOffsets02")
       endif
       noffsets = sum(min(numpixrow,1)) + sum(min(numpixcol,1)) - 1
       ndof = ngoodpix - noffsets
       if (ndof < 1) then
          write(errstr,*) "The number of good pixels in the FOV",  &
               ngoodpix, " is less than the number of offsets", noffsets,  &
               ". The result is not reliable."
          call warning("getOffsets12")
          ndof = 1 
       endif
       averim = sum(real(totrow,kind=double)) / ngoodpix
       write(errstr,*) "Average value over the field of view =", averim
       call message(AppMsg,Verbose)
       where (numpixrow > 4)
          offsetx(0:x_size_window-1,0) = int(real(totrow)/numpixrow  &
                                             -averim/2+0.5,kind=int16)
       endwhere
       where (numpixcol > 4)
          offsety(0:y_size_window-1,0) = int(real(totcol)/numpixcol  &
                                             -averim/2+0.5,kind=int16)
       endwhere

       write(errstr,*) ngoodpix,  &
             " pixels used for computing the offsets in the FOV"
       call message(AppMsg,Verbose)
       write(errstr,*) nbadpix,  &
             " pixels rejected for computing the offsets in the FOV"
       call message(AppMsg,Verbose)
    else   ! For variance
       xrel0 = xref0
       xrel1 = xref1
       yrel0 = yref0
       yrel1 = yref1
    endif

! Compute variance. (nrows+ncol-1) dof must be counted for the offsets

    variance = 0.d0
    do jr = yrel0, yrel1
       do ir = xrel0, xrel1
          if (ccdmap(ir,jr) > 0) then
             variance = variance + (ccdmap(ir,jr) - offsetx(ir+refX0,0)  &
                                    - offsety(jr+refY0,0))**2
          endif
       enddo
    enddo
    variance = variance / ndof
    write(errstr,*) "Variance =", variance
    call message(AppMsg,Verbose)
    call setAttribute(off_tab, "CCDVAR", variance,  &
                      comment="Variance within window")

! Reset initial map

    ccdmap(xref0:xref1,yref0:yref1) = extimg(windX0:windX1,windY0:windY1)

    deallocate(extimg,numpixrow,totrow,numpixcol,totcol)

  end subroutine getOffsets

! Test whether a value is compatible with a distribution

  logical function test_valid(val, border, numsig)

! val    : value to be tested
! border : reference distribution
! numsig : rejection criterion

    integer(int16), intent(in) :: val
    integer, dimension(:), intent(inout) :: border
    real, intent(in) :: numsig

    integer, dimension(:), allocatable :: iperm
    integer :: nborder, ier
    real :: med, sig

    nborder = size(border)
    allocate(iperm(nborder))
    call ipsort(border,nborder,iperm,2,ier)
    if (ier > 0) then
       write(errstr,*) "error in first call to ipsort: ier =", ier
       call error("test_valid01")
    endif
    med = 0.5 * (border(nborder/2+1) + border((nborder-1)/2+1))
    border = int(abs(med-border),kind=int16)
    call ipsort(border,nborder,iperm,2,ier)
    if (ier > 0) then
       write(errstr,*) "error in second call to ipsort: ier =", ier
       call error("test_valid02")
    endif
    sig = 0.5 * (border(nborder/2+1) + border((nborder-1)/2+1))
    test_valid = abs(med-val) <= (sig * numsig)
    deallocate(iperm)

  end function test_valid

! Run the EDU simulator and mask the pixels corresponding to valid events
! Do not mask events with pattern 30 and 31

  subroutine educall(refX0, refY0, event_out, diag_arr, off_tab, eduout)

! ref..     : Window parameters of the input map
! event_out : Name of the output event file
! diag_arr  : Handle to the CCD map array
! off_tab   : Handle to the offset/variance table
! eduout    : List of events (output of EDUSOFT) for THRESH


    character(len=*), intent(in) :: event_out
    type(ArrayT),     intent(in) :: diag_arr
    type(TableT),     intent(in) :: off_tab
    integer,          intent(in) :: refX0, refY0
    type(edu_out),    intent(out):: eduout

    integer(int16), dimension(:,:), allocatable :: image
    integer, dimension(es_nmax) :: evok
    integer :: nbelow, nabove, xrel0, yrel0, xrel1, yrel1
    integer :: nok, cur, ener, x0, y0, x1, y1, pattern
    integer :: edumode = 3   ! Imaging mode

    call message(AppMsg,Verbose,"Simulating EDU analysis ...")

    xrel0 = windX0 - refX0
    xrel1 = windX0 + windDX - 1 - refX0
    yrel0 = windY0 - refY0
    yrel1 = windY0 + windDY - 1 - refY0

! edusoft does not preserve the input map

    allocate(image(windDX,windDY))
    image = ccdmap(xrel0:xrel1,yrel0:yrel1)
    call edusoft(edumode,edupat,windDX,windDY,image,threshold, &
                 offsetx(:,0),offsety(:,0),windX0,windY0,es_nmax,eduout)
    deallocate(image)

! Apply the EMDH thresholds

    nok = 0
    nbelow = 0
    nabove = 0
    do cur = 1, eduout%nevent
       ener = eduout%evt(cur)%e1 + eduout%evt(cur)%e2
       if (ener < emdhlow) then
          nbelow = nbelow + 1
       else if (ener > emdhupp) then
          nabove = nabove + 1
       else
          nok = nok + 1
          evok(nok) = cur
       endif
    enddo
    write(errstr,*) nok, " valid events"
    call message(AppMsg,Verbose)
    write(errstr,*) nbelow, " events below EMDHLOW =", emdhlow
    call message(AppMsg,Verbose)
    write(errstr,*) nabove, " events above EMDHUPP =", emdhupp
    call message(AppMsg,Verbose)

    if (writeevent)  call putevent(eduout, evok, nok, nbelow, nabove,  &
                                   event_out, diag_arr, off_tab)

! Mask the input map (positions are offset by 2 in X and Y)

    do cur = 1, nok
       pattern = eduout%evt(evok(cur))%pattern
       if (pattern < 30) then   ! Do not mask "false" events
          x0 = eduout%evt(evok(cur))%x - 2 - refX0
          x1 = max(x0-1,0)
          y0 = eduout%evt(evok(cur))%y - 2 - refY0
          y1 = max(y0-1,0)
          ccdmap(x1:x0+1,y1:y0+1) = ccdmap(x1:x0+1,y1:y0+1)  &
                                  * (1-patabove(x1-x0:1,y1-y0:1,pattern))
       endif
    enddo

  end subroutine educall

! Mask all pixels above the EDU threshold except those in events below EMDHLOW

  subroutine thresh(refX0, refY0, bright_out, diag_arr, off_tab, eduout)

! ref..      : Window parameters of the input map
! bright_out : Name of the output CCD map file
! diag_arr   : Handle to the CCD map array
! off_tab    : Handle to the offset/variance table
! eduout     : List of events (output of EDUSOFT)

    character(len=*), intent(in) :: bright_out
    type(ArrayT),     intent(in) :: diag_arr
    type(TableT),     intent(in) :: off_tab
    integer,          intent(in) :: refX0, refY0
    type(edu_out),    intent(in) :: eduout

    integer :: xwin0, xwin1, ywin0, ywin1, xref0, xref1, yref0, yref1
    integer :: xrel0, xrel1, yrel0, yrel1
    integer :: numbright
    integer, dimension(es_nmax) :: evbel
    integer :: nbelow, cur, x0, y0, x1, y1, pattern, irow, jrow, n2, nlarge

    type(DataSetT) :: bright_set
    type(ArrayT)   :: bright_arr
    integer(kind=int16), dimension(:,:), pointer :: brightmap

    call message(AppMsg,Verbose,"Masking all pixels above threshold ...")

    xrel0 = windX0 - refX0
    xrel1 = windX0 + windDX - 1 - refX0
    yrel0 = windY0 - refY0
    yrel1 = windY0 + windDY - 1 - refY0

! Remove events recognized by EDUSOFT and below EMDHLOW

    if (onedusoft) then
       nbelow = 0
       nlarge = 0
       do cur = 1, eduout%nevent
          if (eduout%evt(cur)%e1 + eduout%evt(cur)%e2 < emdhlow) then
             pattern = eduout%evt(cur)%pattern
             n2 = sum(patabove(:,:,pattern))
             if (n2 < 3) then   ! Mono and bipixels can be reconstructed
                nbelow = nbelow + 1
                evbel(nbelow) = cur
             else
                nlarge = nlarge + 1
             endif
          endif
       enddo
       if (nlarge > 0) then
          write(errstr,*) nlarge, " events larger than bipixels were below ", &
                          "EMDHLOW =", emdhlow, ". Masked anyway."
          call warning("thresh10")
       endif
       write(errstr,*) nbelow, " events below EMDHLOW =", emdhlow
       call message(AppMsg,Verbose)
       do cur = 1, nbelow
          pattern = eduout%evt(evbel(cur))%pattern
          x0 = eduout%evt(evbel(cur))%x - 2 - refX0
          x1 = max(x0-1,0)
          y0 = eduout%evt(evbel(cur))%y - 2 - refY0
          y1 = max(y0-1,0)
          ccdmap(x1:x0+1,y1:y0+1) = ccdmap(x1:x0+1,y1:y0+1)  &
                                  * (1-patabove(x1-x0:1,y1-y0:1,pattern))
       enddo
    endif

! Subtract the offsets

    do irow = xrel0, xrel1
       ccdmap(irow,yrel0:yrel1) = ccdmap(irow,yrel0:yrel1)  &
                                - offsetx(irow+refX0,0)
    enddo
    do jrow = yrel0, yrel1
       ccdmap(xrel0:xrel1,jrow) = ccdmap(xrel0:xrel1,jrow)  &
                                - offsety(jrow+refY0,0)
    enddo

! Write the bright pixels file

    if (writebright) then
       call message(AppMsg,Verbose,"Writing the map of bright pixels ...")
       bright_set = dataSet(bright_out, Create)
       bright_arr = addArray(bright_set, image_ext, Integer16,  &
                    (/EMOS_MAX_X,EMOS_MAX_Y/), "Map of pixels above threshold")
       call setAttributes(attributable(bright_arr), attributable(diag_arr))
       call setAttribute(bright_arr, "CCDMAP", trim(name(parent(diag_arr))),  &
                         comment="Input diagnostic file")
       call setAttribute(bright_arr, "BRIGHT", 1,  &
                         comment="This is a map of bright pixels")
       call addComment(bright_arr, "Offsets were subtracted from the map")
       brightmap => int16Array2Data(bright_arr)
       brightmap = 0
       xwin0 = max(windX0,dx1) - dx1
       xwin1 = min(windX0+windDX,x_size_window-dx2) - 1 - dx1
       ywin0 = max(windY0,dy1) - dy1
       ywin1 = min(windY0+windDY,y_size_window-dy2) - 1 - dy1
       xref0 = xwin0 + dx1 - refX0
       xref1 = xwin1 + dx1 - refX0
       yref0 = ywin0 + dy1 - refY0
       yref1 = ywin1 + dy1 - refY0
       brightmap(xwin0:xwin1,ywin0:ywin1) = ccdmap(xref0:xref1,yref0:yref1)

! Set all pixels below threshold to 0

       numbright = 0
       do jrow = ywin0, ywin1
          do irow = xwin0, xwin1
             if (brightmap(irow,jrow) < threshold) then
                brightmap(irow,jrow) =  0
             else
                numbright = numbright + 1
             endif
          enddo
       enddo
       write(errstr,*) numbright, " bright pixels in map"
       call message(AppMsg,Verbose)
       call putMapAttributes(bright_arr, xwin0, ywin0, xwin1, ywin1)

! Copy the offsets as an extension

       call message(AppMsg,Verbose,"Writing the offset/variance extension ...")
       call copyBlock(bright_set, block(off_tab), offset_ext)
       call release(bright_set)
    endif

! Set all pixels above threshold to 0 in masked map

    do jrow = yrel0, yrel1
       do irow = xrel0, xrel1
          if (ccdmap(irow,jrow) >= threshold)  ccdmap(irow,jrow) =  &
                             - offsetx(irow+refX0,0) - offsety(jrow+refY0,0)
       enddo
    enddo

! Put the events below EMDHLOW back in

    if (onedusoft) then
       do cur = 1, nbelow
          pattern = eduout%evt(evbel(cur))%pattern
          x0 = eduout%evt(evbel(cur))%x - 2 - refX0
          x1 = max(x0-1,0)
          y0 = eduout%evt(evbel(cur))%y - 2 - refY0
          y1 = max(y0-1,0)
          where (patabove(x1-x0:1,y1-y0:1,pattern) > 0)
             ccdmap(x1:x0+1,y1:y0+1) = eduout%evt(evbel(cur))%e2
          endwhere
          ccdmap(x0,y0) = eduout%evt(evbel(cur))%e1
       enddo
    endif

! Put the offsets back in

    do irow = xrel0, xrel1
       ccdmap(irow,yrel0:yrel1) = ccdmap(irow,yrel0:yrel1)  &
                                + offsetx(irow+refX0,0)
    enddo
    do jrow = yrel0, yrel1
       ccdmap(xrel0:xrel1,jrow) = ccdmap(xrel0:xrel1,jrow)  &
                                + offsety(jrow+refY0,0)
    enddo

  end subroutine thresh

! Read the offsets

  subroutine getoffvar(offvar_in, diag_arr, off_tab)

! offvar_in : Name of the input offset/variance file
! diag_arr  : Handle to the CCD map array
! off_tab   : Handle to the offset/variance table

    character(len=*), intent(in)    :: offvar_in
    type(ArrayT),     intent(in)    :: diag_arr
    type(TableT),     intent(out)   :: off_tab

    type(DataSetT) :: off_set
    type(ColumnT)  :: offx_col, offy_col

    call message(AppMsg,Verbose,"Reading the offsets ...")

    off_set = dataSet(offvar_in, Read)
    off_tab = table(off_set, 0)   ! Extension name is variable
    if (equalKeywords(attributable(diag_arr),  &
                      attributable(off_tab),keycompat)) then
       call message(AppMsg,Verbose,  &
                    "Offset/variance file compatible with event file")
       offx_col = column(off_tab, "COLOFFST", Read)
       offsetx => int16VectorData(offx_col)
       offy_col = column(off_tab, "ROWOFFST", Read)
       offsety => int16VectorData(offy_col)
       if (size(offsetx)<x_size_window.or.size(offsety)<y_size_window) then
          write(errstr,*) "the offsets are incomplete. They have size",  &
             size(offsetx),",",size(offsety)," <",x_size_window,",",  &
             y_size_window," in file ",trim(offvar_in),". File not used."
          call error("getoffvar02")
       endif
    else
       write(errstr,*) "offset/variance file not compatible with DIAGNOSTIC ",&
                       "file. File not used."
       call error("getoffvar03")
    endif

  end subroutine getoffvar

! Read the input CCD map

  subroutine readmap(ccd_in, diag_arr, refX0, refY0)

! ccd_in   : Name of the input CCD map file
! diag_arr : Handle to the CCD map array
! ref..    : Window parameters of the input map

    character(len=*), intent(in)  :: ccd_in
    type(ArrayT),     intent(out) :: diag_arr
    integer,          intent(out) :: refX0, refY0

    type(DataSetT) :: diag_set
    integer, dimension(:), pointer :: sizarray
    logical :: fromodf
    integer :: inX1, inY1, windX1, windY1, inDX, inDY, inX0, inY0

    call message(AppMsg,Verbose,"Reading the input map ...")

    diag_set = dataSet(ccd_in, Read)
    fromodf = .not. hasBlock(diag_set, image_ext)
    if (fromodf) then
       call message(AppMsg,Verbose,  &
                    "The input CCD map is straight from the ODF")
       diag_arr = array(diag_set, 0, Read)   ! Extension name is variable
    else
       call message(AppMsg,Verbose,"The input CCD map is a masked image")
       diag_arr = array(diag_set, image_ext, Read)
       if (.not. wasDone(block(diag_arr),"MASKED")) then
          write(errstr,*) "The input image is not straight from the ODF ",  &
             "but is not the result of a previous call to EMDIAG. ",  &
             "Proceeding anyway."
          call message(AppMsg,Sparse)
       endif
    endif

    call initCAL(diag_arr)

    if (numberOfDimensions(diag_arr) == 2) then
       sizarray => dimensions(diag_arr)
       inDX = sizarray(0)
       inDY = sizarray(1)
       if (max(inDX-x_size_window,inDY-y_size_window) > 0) then
! Should try to make emdiag more general (array of any size)
          write(errstr,*) "The array ", trim(qualifiedName(diag_arr)),  &
                          " has dimensions", sizarray,  &
                          " larger than", x_size_window, y_size_window
          call error("readmap03")
       else
          ccdmap => int16Array2Data(diag_arr)
       endif
    else
       write(errstr,*) "The array ", trim(qualifiedName(diag_arr)) ," of ",  &
                       trim(ccd_in), " is not 2-D."
       call error("readmap02")
    endif

    if (fromodf) then
       if (hasAttribute(diag_arr,"WINDOWDX")) then
          if (inDX /= int32Attribute(diag_arr,"WINDOWDX")) then
             write(errstr,*) "WINDOWDX =",int32Attribute(diag_arr,"WINDOWDX"),&
                " in input map is not compatible with array size", inDX
             call error("readmap04")
          endif
       endif
       if (hasAttribute(diag_arr,"WINDOWDY")) then
          if (inDY /= int32Attribute(diag_arr,"WINDOWDY")) then
             write(errstr,*) "WINDOWDY =",int32Attribute(diag_arr,"WINDOWDY"),&
                " in input map is not compatible with array size", inDY
             call error("readmap05")
          endif
       endif
       if (hasAttribute(diag_arr,"WINDOWX0")) then
          refX0 = int32Attribute(diag_arr,"WINDOWX0")
       else if (inDX == x_size_window) then
          refX0 = 0
       else
          call error("readmap06", "WINDOWX0 is not present in the input map.")
       endif
       if (hasAttribute(diag_arr,"WINDOWY0")) then
          refY0 = int32Attribute(diag_arr,"WINDOWY0")
       else if (inDY == y_size_window) then
          refY0 = 0
       else
          call error("readmap07", "WINDOWY0 is not present in the input map.")
       endif
       inX0 = refX0
       inY0 = refY0
       if (hasAttribute(diag_arr,"EDUTHR") .and.  &
           hasAttribute(diag_arr,"NPIXEL")) then
          if (int32Attribute(diag_arr,"EDUTHR") == 0 .and.  &
              int32Attribute(diag_arr,"NPIXEL") /= inDX*inDY) then
             write(errstr,*) "NPIXEL =", int32Attribute(diag_arr,"NPIXEL"),  &
                " is not equal to the total window size", inDX*inDY,  &
                " in the input file. Proceeding anyway."
          call warning("readmap11")
          endif
       endif
    else if ((inDX+dx1+dx2 == x_size_window) .and.  &
             (inDY+dy1+dy2 == y_size_window)) then   ! Masked CCD map

! In that case the actual window parameters may not be the image size

       refX0 = dx1
       refY0 = dy1
       if (hasAttribute(diag_arr,"WINDOWX0")) then
          inX0 = int32Attribute(diag_arr,"WINDOWX0") + dx1 - 1
       else
          inX0 = refX0
       endif
       if (hasAttribute(diag_arr,"WINDOWY0")) then
          inY0 = int32Attribute(diag_arr,"WINDOWY0") + dy1 - 1
       else
          inY0 = refY0
       endif
       if (hasAttribute(diag_arr,"WINDOWDX"))  &
          inDX = int32Attribute(diag_arr,"WINDOWDX")
       if (hasAttribute(diag_arr,"WINDOWDY"))  &
          inDY = int32Attribute(diag_arr,"WINDOWDY")

! Get the thresholds if they were not set manually

       if (threshold <= 0 .and. hasAttribute(diag_arr,"EDUTHR")) then
          threshold = int16Attribute(diag_arr,"EDUTHR")
          call message(AppMsg, Verbose,  &
                       "The EDU threshold was taken from the input file")
       endif
       if (emdhlow <= 0 .and. hasAttribute(diag_arr,"EMDHLOW")) then
          emdhlow = int32Attribute(diag_arr,"EMDHLOW")
          call message(AppMsg, Verbose,  &
                      "The EMDH lower threshold was taken from the input file")
       endif
    else
       write(errstr,*) "The input map ", trim(ccd_in), " doesn't have ",  &
          "the expected size", x_size_window-dx1-dx2, y_size_window-dy1-dy2 
       call error("readmap08")
    endif

    if (windX0 == 1 .and. windDX == 600) then
       windX0 = 0                    ! Full frame mode
       windDX = windDX + dx1 + dx2
    else
       windX0 = windX0 + dx1 - 1     ! Window mode
    endif
    if (windY0 == 1 .and. windDY == 600) then
       windY0 = 0                    ! Full frame mode
       windDY = windDY + dy1 + dy2
    else
       windY0 = windY0 + dy1 - 1     ! Window mode
    endif
    windX1 = windX0 + windDX
    windY1 = windY0 + windDY
    inX1  = inX0  + inDX
    inY1  = inY0  + inDY
    windX0 = max(windX0,inX0)
    windY0 = max(windY0,inY0)
    windDX = min(windX1,inX1) - windX0
    windDY = min(windY1,inY1) - windY0
    if (min(windDX,windDY) < 1) then
       write(errstr,*) "The intersection of the actual window in the data [", &
                       inX0-dx1+1, "-", inX1-dx1, "] x [", inY0-dy1+1,  &
                       "-", inY1-dy1, "] with the required window is empty."
       call error("readmap09")
    endif
    write(errstr,*) "Window parameters for the EDU: X0 =", windX0,  &
                    "   Y0 =", windY0
    call message(AppMsg,Verbose)
    write(errstr,*) "Window parameters for the EDU: DX =", windDX,  &
                    "   DY =", windDY
    call message(AppMsg,Verbose)

    if (onedusoft .or. onthresh) then
       write(errstr,*) "EDU threshold set at", threshold
       call message(AppMsg,Sparse)
    endif
    if (onedusoft) then
       if (emdhlow > threshold) then
          write(errstr,*) "Lower EMDH threshold set at", emdhlow
          call message(AppMsg,Sparse)
       else
          emdhlow = threshold
       endif
       if (emdhupp < emdhlow) then
          write(errstr,*) "EMDHup =", emdhupp, " should be larger than ",  &
             "EMDHlow =", emdhlow, "for EDUSOFT to find any events. ",  &
             "EDUSOFT not done."
          call warning("readmap12")
          onedusoft = .false.
       else
          write(errstr,*) "Upper EMDH threshold set at", emdhupp
          call message(AppMsg,Sparse)
       endif
    endif

  end subroutine readmap

! Write the masked CCD map

  subroutine putmap(refX0, refY0, ccd_out, diag_arr, off_tab)

! ccd_out  : Name of the output CCD map file
! diag_arr : Handle to the CCD map array
! off_tab  : Handle to the offset/variance table
! ref..    : Window parameters of the input map

    character(len=*), intent(in) :: ccd_out
    type(ArrayT),     intent(in) :: diag_arr
    type(TableT),     intent(in) :: off_tab
    integer,          intent(in) :: refX0, refY0

    integer :: xwin0, xwin1, ywin0, ywin1, xrel0, xrel1, yrel0, yrel1

    type(DataSetT) :: masked_set
    type(ArrayT)   :: masked_arr
    integer(kind=int16), dimension(:,:), pointer :: maskedmap

    call message(AppMsg,Verbose,"Writing the masked CCD map ...")

    masked_set = dataSet(ccd_out, Create)
    masked_arr = addArray(masked_set, image_ext, Integer16,  &
                          (/EMOS_MAX_X,EMOS_MAX_Y/), "Masked CCD map")
    call setAttributes(attributable(masked_arr), attributable(diag_arr))
    call setAttribute(masked_arr, "CCDMAP", trim(name(parent(diag_arr))),  &
                      comment="Input diagnostic file")
    call setAttribute(masked_arr, "MASKED", 1,  &
                      comment="This is a masked CCD map")
    maskedmap => int16Array2Data(masked_arr)
    maskedmap = 0

    xwin0 = max(windX0,dx1) - dx1
    xwin1 = min(windX0+windDX,x_size_window-dx2) - 1 - dx1
    ywin0 = max(windY0,dy1) - dy1
    ywin1 = min(windY0+windDY,y_size_window-dy2) - 1 - dy1
    xrel0 = xwin0 + dx1 - refX0
    xrel1 = xwin1 + dx1 - refX0
    yrel0 = ywin0 + dy1 - refY0
    yrel1 = ywin1 + dy1 - refY0
    maskedmap(xwin0:xwin1,ywin0:ywin1) = ccdmap(xrel0:xrel1,yrel0:yrel1)

    call putMapAttributes(masked_arr, xwin0, ywin0, xwin1, ywin1)

! Copy the offsets as an extension

    call message(AppMsg,Verbose,"Writing the offset/variance extension ...")
    call copyBlock(masked_set, block(off_tab), offset_ext)

    call release(masked_set)

  end subroutine putmap

! Write the event list

  subroutine putevent(eduout, evok, nok, nbelow, nabove,  &
                      event_out, diag_arr, off_tab)

! eduout    : Structure in output of EDUSOFT
! evok      : Indices of events between the EMDH thresholds
! nok       : Number of such valid events
! nbelow    : Number of events below lower EMDH threshold
! nabove    : Number of events above upper EMDH threshold
! event_out : Name of the output event file
! diag_arr  : Handle to the CCD map array
! off_tab   : Handle to the offset/variance table

    type(edu_out),    intent(in) :: eduout
    character(len=*), intent(in) :: event_out
    type(ArrayT),     intent(in) :: diag_arr
    type(TableT),     intent(in) :: off_tab
    integer,          intent(in) :: nbelow, nabove, nok
    integer, dimension(nok), intent(in) :: evok

    type(DataSetT) :: ev_set
    type(TableT)   :: ev_tab
    type(ColumnT)  :: fram_col, rawx_col, rawy_col, patt_col, peri_col, &
                      e1_col, e2_col, e3_col, e4_col
    integer(kind=int8),  dimension(:), pointer :: pattern, peripix
    integer(kind=int16), dimension(:), pointer :: rawx, rawy, e1, e2, e3, e4
    integer(kind=int32), dimension(:), pointer :: frame

    call message(AppMsg,Verbose,"Writing the event list ...")

    if (nok > eduout%nevent) then
       write(errstr,*) "Number of valid events", nok,  &
          " larger than the total number of events", eduout%nevent
       call error("putevent01")
    endif
    if (maxval(evok) > eduout%nevent) then
       write(errstr,*) "At least one valid event has index", maxval(evok),  &
          " larger than the total number of events", eduout%nevent
       call error("putevent02")
    endif

    ev_set = dataSet(event_out, Create)
    ev_tab = addTable(ev_set, event_ext, nok, "Event list")
    call setAttributes(attributable(ev_tab), attributable(diag_arr))
    call setAttribute(ev_tab, "DATATYPE", "IMAGE.EL")
    call setAttribute(ev_tab, "CCDMAP", trim(name(parent(diag_arr))),  &
                      comment="Input diagnostic file")
    call keywordDone(block(ev_tab), taskname, edusoft_on)
    call setAttribute(ev_tab, "WINDOWX0", windX0,  &
                      comment="X-coordinate of bottom left corner of window")
    call setAttribute(ev_tab, "WINDOWY0", windY0,  &
                      comment="Y-coordinate of bottom left corner of window")
    call setAttribute(ev_tab, "WINDOWDX", windDX,  &
                      comment="Size, along X-axis, of window")
    call setAttribute(ev_tab, "WINDOWDY", windDY,  &
                      comment="Size, along Y-axis, of window")
    if (onoffset)  &
       call addComment(ev_tab, "Offsets computed on the map itself")
    call setAttribute(ev_tab, "EDUTHR", threshold,  &
                      comment="EDU threshold used for event recognition")
    call setAttribute(ev_tab, "EMDHLOW", emdhlow,  &
                     comment="EMDH lower threshold used for event recognition")
    call setAttribute(ev_tab, "EMDHUPP", emdhupp,  &
                   comment="EMDH upper threshold used for event recognition")
    call setAttribute(ev_tab, "NPIXEL", eduout%npix,  &
                      comment="Number of pixels above EDU threshold in window")
    call setAttribute(ev_tab, "NVALID", nok,  &
                      comment="Number of valid events")
    call setAttribute(ev_tab, "NBELOW", nbelow,  &
                      comment="Number of events below lower EMDH threshold")
    call setAttribute(ev_tab, "NABOVE", nabove,  &
                      comment="Number of events above upper EMDH threshold")

    fram_col =  addColumn(ev_tab, "FRAME",    Integer32, "COUNTER",  &
                          comment="Frame number")
    frame    => int32Data(fram_col)
    frame    =  int32Attribute(diag_arr,"FRAME")
    call release(fram_col)
    rawx_col =  addColumn(ev_tab, "RAWX",     Integer16, "PIXEL",  &
                          comment="Event X position")
    rawx     => int16Data(rawx_col)
    rawx     =  eduout%evt(evok)%x
    call release(fram_col)
    rawy_col =  addColumn(ev_tab, "RAWY",     Integer16, "PIXEL",  &
                          comment="Event Y position")
    rawy     => int16Data(rawy_col)
    rawy     =  eduout%evt(evok)%y
    call release(fram_col)
    e1_col   =  addColumn(ev_tab, "ENERGYE1", Integer16, "CHANNEL",  &
                          comment="Event energy (E1)")
    e1       => int16Data(e1_col)
    e1       =  eduout%evt(evok)%e1
    call release(e1_col)
    e2_col   =  addColumn(ev_tab, "ENERGYE2", Integer16, "CHANNEL",  &
                          comment="Event energy (E2)")
    e2       => int16Data(e2_col)
    e2       =  eduout%evt(evok)%e2
    call release(e2_col)
    e3_col   =  addColumn(ev_tab, "ENERGYE3", Integer16, "CHANNEL",  &
                          comment="Event energy (E3)")
    e3       => int16Data(e3_col)
    e3       =  eduout%evt(evok)%e3
    call release(e3_col)
    e4_col   =  addColumn(ev_tab, "ENERGYE4", Integer16, "CHANNEL",  &
                          comment="Event energy (E4)")
    e4       => int16Data(e4_col)
    e4       =  eduout%evt(evok)%e4
    call release(e4_col)
    patt_col =  addColumn(ev_tab, "PATTERN",  Integer8,  &
                          comment="Pattern number")
    pattern  => int8Data(patt_col)
    pattern  =  eduout%evt(evok)%pattern
    call release(patt_col)
    peri_col =  addColumn(ev_tab, "PERIPIX",  Integer8,  &
                          comment="Peripheral pixels above threshold")
    peripix  => int8Data(peri_col)
    peripix  =  eduout%evt(evok)%peripix
    call release(peri_col)

    call release(ev_tab)

! Copy the offsets as an extension

    call message(AppMsg,Verbose,"Writing the offset/variance extension ...")
    call copyBlock(ev_set, block(off_tab), offset_ext)

    call release(ev_set)

  end subroutine putevent

! Copy keywords from array extension to binary table

! Write attributes on the output array and release it

  subroutine putMapAttributes(masked_arr, xwin0, ywin0, xwin1, ywin1)

! masked_arr : Handle to the map array
! ..win..    : Window parameters of the output map

    type(ArrayT),     intent(in) :: masked_arr
    integer,          intent(in) :: xwin0, ywin0, xwin1, ywin1

    call setAttribute(masked_arr, "WINDOWX0", xwin0+1,  &
                      comment="X-coordinate of bottom left corner of window")
    call setAttribute(masked_arr, "WINDOWY0", ywin0+1,  &
                      comment="Y-coordinate of bottom left corner of window")
    call setAttribute(masked_arr, "WINDOWDX", xwin1-xwin0+1,  &
                      comment="Size, along X-axis, of window")
    call setAttribute(masked_arr, "WINDOWDY", ywin1-ywin0+1,  &
                      comment="Size, along Y-axis, of window")
    if (onoffset)  &
       call addComment(masked_arr, "Offsets computed on the map itself")
    call setAttribute(masked_arr, "EDUTHR", threshold,  &
                      comment="EDU threshold used for event recognition")
    call setAttribute(masked_arr, "EMDHLOW", emdhlow,  &
                     comment="EMDH lower threshold used for event recognition")
    if (onedusoft) then
       call setAttribute(masked_arr, "EMDHUPP", emdhupp,  &
                   comment="EMDH upper threshold used for event recognition")
       call keywordDone(block(masked_arr), taskname, edusoft_on)
    endif
    if (onthresh)  call keywordDone(block(masked_arr), taskname, thresh_on)
    call release(masked_arr)

  end subroutine putMapAttributes

! Parameter reading

  subroutine getparamvalues(ccd_in, ccd_out, offvarName, event_out, bright_out)

! ccd_in     : Name of the input CCD map file
! ccd_out    : Name of the output CCD map file
! offvarName : Name of the input or output offset/variance file
! event_out  : Name of the output event list file
! bright_out : Name of the output CCD map file for bright pixels

    character(len=*), intent(out) :: event_out, offvarName,  &
                                     ccd_in, ccd_out, bright_out

    call message(AppMsg,Verbose,"Reading the parameters ...")

    ccd_in = stringParameter(inccd)
    if (len_trim(ccd_in) == 0) &
       call error("getparamvalues01","Blank input CCD map file name")
    write(errstr,*) "Input CCD map file : ",trim(ccd_in)
    call message(AppMsg,Sparse)

    windX0 = intParameter(windowX0)
    windDX = intParameter(windowDX)
    windY0 = intParameter(windowY0)
    windDY = intParameter(windowDY)

    onoffset  = stringParameter(offset_on) == 'compute'
    if (onoffset) then
       call message(AppMsg,Sparse,"OFFSET  will be applied")
       numsig = realParameter(ksig)
       write(errstr,*) "Threshold for pixel rejection in OFFSET set at",  &
                       numsig, " sigmas."
       call message(AppMsg,Sparse)
       onsubbad = booleanParameter(subbad)
       if (onsubbad) then
           write(errstr,*) "The subtraction of bad pixels from the CCD ",  &
                           "is not implemented yet."
           call warning("getparamvalues99")
!          write(errstr,*) "Bad pixels in the CCF will be masked out ",  &
!                          "when calculating the offsets"
!          call message(AppMsg,Sparse)
       endif
       onselfov = booleanParameter(selfov)
       if (onselfov) then
          write(errstr,*) "The pixels outside the field of view will be ",  &
                          "masked out when calculating the offsets"
          call message(AppMsg,Sparse)
       endif
       writeoffvar = booleanParameter(writeoff)
       if (writeoffvar) then
          offvarName = stringParameter(outoffvar)
          write(errstr,*) "Output offset/variance file : ",trim(offvarName)
          call message(AppMsg,Sparse)
       endif
    else
       offvarName = stringParameter(inoffvar)
    endif

    onedusoft = booleanParameter(edusoft_on)
    onthresh  = booleanParameter(thresh_on)
    if (onedusoft) call message(AppMsg,Sparse,"EDUSOFT will be applied")
    if (onthresh)  call message(AppMsg,Sparse,"THRESH  will be applied")

    writemap = .false.
    if (onedusoft .or. onthresh) then
       threshold = intParameter(EDUthresh)
       emdhlow = intParameter(EMDHlower)
       writemap = booleanParameter(writemasked)
       if (onoffset) then
          call message(AppMsg,Sparse,  &
                       "Offsets will be derived from the map itself.")
       else
          write(errstr,*) "Offsets will be taken from file : ",trim(offvarName)
          call message(AppMsg,Sparse)
       endif
    endif

    writeevent = .false.
    if (onedusoft) then
       emdhupp = intParameter(EMDHupper)
       writeevent = booleanParameter(writeevt)
    endif

    if (writemap) then
       ccd_out = stringParameter(maskedfile)
       write(errstr,*) "Output masked CCD map file : ",trim(ccd_out)
       call message(AppMsg,Sparse)
    endif
    if (writeevent) then
       event_out = stringParameter(outevent)
       write(errstr,*) "Output event file : ",trim(event_out)
       call message(AppMsg,Sparse)
    endif

    if (onthresh) then
       writebright = booleanParameter(writebri)
       if (writebright) then
          bright_out = stringParameter(brightfile)
          write(errstr,*) "Output CCD map file for bright pixels: ",  &
                          trim(bright_out)
          call message(AppMsg,Sparse)
       endif
    endif

  end subroutine getparamvalues

! Simulates setStateFromScienceFile for EPIC MOS for arrays
! Gets patterns from CAL

  subroutine initCAL(diag_arr)

! diag_arr    : handle to the input array

    type(ArrayT), intent(in)       :: diag_arr
    integer(kind=int8), dimension(:,:,:), pointer :: patterns
    integer(kind=int16) :: eduThreshold
    type(ModeParameterDataType)  :: modeParameters

    integer :: datamode, jr, npattern, frmtime

    call CAL_setState(block(diag_arr))

! Artificially change the mode so that the pattern library may be retrieved
! Timing mode may occur only if FRMTIME=0 and WINDOWDY=602 or 1024
! The reciprocal is not completely true (one may use free run with 602 lines)

    frmtime = int32Attribute(diag_arr,"FRMTIME")
    if (int32Attribute(diag_arr,"WINDOWDY") >= y_size_window .and.  &
        (frmtime == 0 .or. frmtime == 1024)) then
       datamode = TIMING
    else
       datamode = IMAGING
    endif
    call CAL_setState(ccdModeId=datamode)

! Get under/overscans

    call CAL_getModeParameters(modeParameters)
    dx1 = modeParameters%underscanX
    dx2 = modeParameters%overscanX
    dy1 = modeParameters%underscanY
    dy2 = modeParameters%overscanY
    x_size_window = EMOS_MAX_X + dx1 + dx2
    y_size_window = EMOS_MAX_Y + dy1 + dy2

! Initialize patterns for EDUSOFT
! Transpose the pattern library and set the 2s to 0

    if (onedusoft .or. onthresh) then
       call CAL_getEventPatterns(patterns, eduThreshold)
       call pat_init(patterns,edupat)
       do npattern = 0,edu_npat-1
          do jr = -1,1
             patabove(:,jr,npattern) = patterns(npattern,jr+2,1:3)
          enddo
       enddo
       where (patabove > 1)
          patabove = 0
       endwhere
    endif

  end subroutine initCAL

end module emdiag_mod
